#!/usr/bin/perl

use strict;
use warnings;

use CGI::Pretty;
use CGI::Carp;

sub init;
sub dispatch_to_context;

sub main_menu;

sub get_players_struct;
sub get_games_struct;

sub add_player_form;
sub add_player;
sub add_player_fail($);

sub player_list;
sub game_list;

sub enter_game_result_form;
sub enter_game_result;
sub enter_game_fail($);

sub lookup_player;
sub lookup_player_form;

sub modify_player;
sub modify_player_form;
sub modify_player_disc;

sub check_game;
sub check_game_form;

sub rank_format_data;
sub ladder_rank_format_human;
sub strength_rank_format_human;

sub help;

sub start_html;

sub get_go_ranks;

sub translate_rank($);

use vars qw(@ladder_ranks $epsilon);


# main()

my $cgi = new CGI;

my $cgi_filename = "go_ladder.cgi";

init();

dispatch_to_context();

1;


# init()

# Initialise

sub init() {

  # setup fp epsilon

  $epsilon = 0.0001;

  # build ladder ranks

  # 25 - 20 kyu / 1

  for (my $i = 30; $i > 19; $i--) {
	push(@ladder_ranks, $i);
  }

  # 19 - 10 kyu / 0.5

  for (my $i = 19; $i > 9.5; $i -= 0.5) {
	push(@ladder_ranks, $i);
  }

  # 9.5 - 1 dan / 0.2

  for (my $i = 9.5; $i > 0; $i -= 0.2) {
	push(@ladder_ranks, $i);
  }

  # 1 dan - 7 dan / 0.1

  for (my $i = 0; $i > -6; $i -= 0.1) {
	push(@ladder_ranks, $i);
  }

  foreach my $j (@ladder_ranks) {
	#		print $j . "\t";
	#	print ladder_rank_format_human($j) . "\n";
  }
}



# dispatch_to_context()

# Dispatch to context

sub dispatch_to_context {

  if ($cgi->param('main_menu')) {
	main_menu();
  } elsif ($cgi->param('player_list')) {
	player_list();
  } elsif ($cgi->param('game_list')) {
	game_list();
  } elsif ($cgi->param('add_player_form')) {
	add_player_form();
  } elsif ($cgi->param('add_player')) {
	add_player();
  } elsif ($cgi->param('enter_game_result_form')) {
	enter_game_result_form();
  } elsif ($cgi->param('enter_game_result')) {
	enter_game_result();
  } elsif ($cgi->param('check_game_form')) {
	check_game_form();
  } elsif ($cgi->param('check_game')) {
	check_game();
  } elsif ($cgi->param('lookup_player_form')) {
	lookup_player_form();
  } elsif ($cgi->param('lookup_player')) {
	lookup_player();
  } elsif ($cgi->param('modify_player_disc')) {
	modify_player_disc();
  } elsif ($cgi->param('modify_player_form')) {
	modify_player_form();
  } elsif ($cgi->param('modify_player')) {
	modify_player();
  } elsif ($cgi->param('help')) {
	help();
  } else {
	main_menu();
  }
}


# main_menu()

# Draw the main menu

sub main_menu {

  start_html('CUGOS Ladder Main Menu');

  print $cgi->h1("CUGOS Ladder Main Menu");
  print $cgi->p("Welcome to the CUGOS Ladder system.");
  print $cgi->p("Make your choice...");
	
  # main menu form

  print $cgi->startform("POST", "$cgi_filename");

  my @rows;

  push(@rows, $cgi->td({-class=>'main_menu'}, [$cgi->submit(-class=>'main_menu', -name=>"player_list", -value=>"Ladder / Player List"),
											   $cgi->submit(-class=>'main_menu', -name=>"game_list", -value=>"Game List")]));

  push(@rows, $cgi->td({-class=>'main_menu'}));

  push(@rows, $cgi->td({-class=>'main_menu'}, [$cgi->submit(-class=>'main_menu', -name=>"check_game_form", -value=>"Check Game Handicap"), 
											   $cgi->submit(-class=>'main_menu', -name=>"enter_game_result_form", -value=>"Enter Game Result")]));

  push(@rows, $cgi->td({-class=>'main_menu'}, [$cgi->submit(-class=>'main_menu', -name=>"add_player_form", -value=>"Add Player "),
											   $cgi->submit(-class=>'main_menu', -name=>"lookup_player_form", -value=>"Lookup Player")]));

  push(@rows, $cgi->td({-class=>'main_menu'}, [$cgi->submit(-class=>'main_menu', -name=>"modify_player_form", -value=>"Modify Player ")]));

  push(@rows, $cgi->td({-class=>'main_menu'}));
	
  push(@rows, $cgi->td({-class=>'main_menu'}, $cgi->submit(-class=>'main_menu', -name=>"help", -value=>"Help / Ladder Rules")));

  print $cgi->table({-class=>'main_menu'}, "", "", $cgi->Tr({-class=>'main_menu'},\@rows));
	
  print $cgi->endform();
  print $cgi->end_html();
}


# player_list()

# Display player list

sub player_list {

  my %players = get_players_struct();
	
  # Sort on ladder rank

  my @players_list = sort { $$a{"ladder_rank"} <=> $$b{"ladder_rank"} } values %players;
	
  # Print HTML

  start_html('CUGOS Ladder Player List');
  print $cgi->h1("CUGOS Ladder Player List");

  return_main_menu_html();
	
  # player list

  my @rows = $cgi->th({-class=>'player_list'}, ["Position", "Name", "Email", "Strength"]);
	
  foreach my $player_ref (@players_list) {
		
	push(@rows, $cgi->td({-class=>'player_list'},
						 [$$player_ref{"ladder_rank"}, $$player_ref{"player_name"}, $$player_ref{"email"}, $$player_ref{"strength"}]));
  }

  print $cgi->table({class=>'player_list'}, "", "", $cgi->Tr(\@rows));

  return_main_menu_html();

  print $cgi->end_html();
}

# game_list()

# Display games list

sub game_list {

  my @games = get_games_struct();
	
  # Print HTML

  start_html('CUGOS Ladder Game List');
  print $cgi->h1("CUGOS Ladder Game List");
	
  # games list

  my @rows = $cgi->th({-class=>'game_list'}, ["Date", "Black", "", "White"]);
	
  foreach my $game_ref (@games) {
		
	my %game_data = %$game_ref;

	my $result_string;

	if ($game_data{"result"}) {
	  $result_string = " beat ";
	} else {
	  $result_string = " lost to ";
	}

    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($game_data{"time"});
	my $time_str = "$mday" . "/" . ($mon + 1) . "/" . ($year + 1900);

	push(@rows, $cgi->td({-class=>'game_list'}, [$time_str, $game_data{"white"}, $result_string, $game_data{"black"}]));
  }

  print $cgi->table({-class=>'game_list'}, "", $cgi->caption({-class=>'game_list'}, "Game List"), $cgi->Tr({-class=>'game_list'}, \@rows));
  
  return_main_menu_html();
  
  print $cgi->end_html();
}

# Player data format is:
# (name) (email) (strength) (ladder position)

sub get_players_struct {

  open(PLAYERS, "go_data/players.txt") or die "can't find players.txt $!";

  my %players;
	
  # Load players
	
  while (defined(my $player_str = <PLAYERS>)) {
		
	my %player_data;
		
	if ($player_str =~ /([\S\ ]+) \t+ ([\S]*) \t+ ([\d\ \w]+) \t+ ([\d]+)/x) {
			
	  $player_data{"player_name"} = $1;
	  $player_data{"email"} = $2;
	  $player_data{"strength"} = $3;
	  $player_data{"ladder_rank"} = $4;

	  $players{$player_data{"player_name"}} = \%player_data;
	} else {
	  warn("Invalid data in players.txt");
	}
  }

  close(PLAYERS);

  return %players;
}

sub get_games_struct {

  open(GAMES, "go_data/games.txt") or die "can't find games.txt $!";

  my @games;
	
  # Load games
	
  while (defined(my $games_str = <GAMES>)) {
		
	my %game_data;
		
	if ($games_str =~ /([\S\ ]+) \t+ ([\S\ ]+) \t+ ([\d]+) \t+ ([\d]+)/x) {
			
	  $game_data{"white"} = $1;
	  $game_data{"black"} = $2;
	  $game_data{"result"} = $3;
	  $game_data{"time"} = $4;
	  push(@games, \%game_data);
	} else {
	  warn("Invalid data in games.txt");
	}
  }

  close(GAMES);

  return @games;
}


sub add_player_form {

  # Draw add player form

  start_html('CUGOS Ladder Add Player');
  print $cgi->h1("CUGOS Ladder Add Player");
  print $cgi->p("Fill in the form below to add a new player...");
	
  # main menu form

  print $cgi->startform("POST", "$cgi_filename");

  print $cgi->p("Player name: ", 
				$cgi->textfield('Player name', '', 20, 30));

  print $cgi->p("Email (optional): ", 
				$cgi->textfield('Player email', '', 20, 30));

  my @go_ranks = get_go_ranks();

  print $cgi->p("Player strength: ",
				$cgi->popup_menu(-name=>'Player strength', -values=>\@go_ranks));

  print $cgi->p($cgi->submit("add_player", "Add Player"));

  print $cgi->endform();

  return_main_menu_html();
	
  print $cgi->end_html();
}

sub add_player {

  # Check the name is unique

  my $player_name = $cgi->param("Player name");

  my %players = get_players_struct();

  if (exists $players{$player_name}) {
	add_player_fail("Name is not unique");
	return;
  }

  # Create new record
	
  my $player_strength = $cgi->param("Player strength");
  my $player_email = $cgi->param("Player email");

  # Check for valid entries
	
  # empty entries

  if (!($cgi->param("Player name"))) {
	add_player_fail("Empty entries");
	return;
  }

  # Calculate ladder position

  my $ladder_ranks = 0;

  foreach my $player_ref (values %players) {
	$ladder_ranks++;
  }

  my $player_ladder = $ladder_ranks + 1;

  # Append data

  # Open players file for appending

  open(PLAYERS, ">> go_data/players.txt") or die "can't find players.txt $!";

  print PLAYERS $player_name . "\t" . $player_email . "\t" . $player_strength . "\t" . $player_ladder . "\n";

  # Print success HTML

  start_html('CUGOS Ladder Player Add Succeeded!');
  print $cgi->h1("CUGOS Ladder Player Add Succeeded!");
	
  print $cgi->p("Player ", $player_name, " club rank ", $player_strength, " successfully added to the ladder at ladder rank ", $player_ladder, ".\n");

  return_main_menu_html();

  print $cgi->end_html();

  close PLAYERS;
}

sub add_player_fail($) {

  my ($error_message) = @_;

  # Print failure HTML

  start_html('CUGOS Ladder Player Add Failed!');
  print $cgi->h1("CUGOS Ladder Player Add Failed!");
	
  print $cgi->p($error_message);
  print $cgi->p("Please re-enter data and try again");

  print $cgi->startform("POST", "$cgi_filename");
  print $cgi->p($cgi->submit("add_player_form", "Add Player"));

  return_main_menu_html();

  print $cgi->end_html();
}

sub enter_game_result_form {
	
  my %players = get_players_struct();

  # build players names array

  my @player_names;

  foreach my $player_ref (values(%players)) {
		
	push(@player_names, $$player_ref{"player_name"});
  }

  @player_names = sort @player_names;

  # Draw enter game form

  start_html('CUGOS Ladder Enter Game Result');
  print $cgi->h1("CUGOS Ladder Add Player");
  print $cgi->p("Fill in the form below to enter a game result...");
	
	
  print $cgi->startform("POST", "$cgi_filename");

  print "<p>";
  print "Black: ";
  print $cgi->popup_menu(-name=>'Black',
						 -values=>\@player_names);
	
  my @outcomes = ("beat", "lost to");

  print $cgi->popup_menu(-name=>'Result',
						 -values=>\@outcomes);

  print " White: ";
  print $cgi->popup_menu(-name=>'White',
						 -values=>\@player_names);

  print "</p>";

  print "<p>";
  print "Check here the difference between ladder grades is > 3 but this game should be accepted:";

  print $cgi->checkbox(-name=>'Confirm', -value=>'on', -label=>'');

  print $cgi->p($cgi->submit("enter_game_result", "Enter Game Result"));
  print $cgi->endform();

  return_main_menu_html();
	
  print $cgi->end_html();
}

sub enter_game_result() {

  # Open players file for appending

  open(GAMES, ">> go_data/games.txt") or die "can't find games.txt $!";

  # Create new record
   
  my $black_name = $cgi->param("Black");
  my $white_name = $cgi->param("White");
  my $game_result = 0;

  if ($cgi->param("Result") eq "beat") {
	$game_result = 1;
  }

  # Check person hasn't beaten themself
	
  if ($cgi->param("Black") eq $cgi->param("White")) {
	enter_game_fail("That must have been an <i>interesting</i> game.");
	return;
  }

  # Find players

  my %players = get_players_struct();

  my $black_ref = $players{$black_name};
  my $white_ref = $players{$white_name};

  # Check difference in rank isn't > 3

  if (abs($$black_ref{"ladder_rank"} - $$white_ref{"ladder_rank"}) > 3) {
	if ($cgi->param("Confirm") ne "on") {
	  enter_game_fail("Ladder rank difference between players is > 3. Please recheck your entry");
	  print $cgi->param("Confirm");
	  return;
	}
  }

  # Append data to games file

  print GAMES $black_name . "\t" . $white_name . "\t" . $game_result . "\t" . time . "\n";

  # Promote/Demote players

  # Calculate new ranks

  my $winner_ref;
  my $loser_ref;

  if ($game_result) {			# black victory
	
	$winner_ref = $black_ref;
	$loser_ref = $white_ref;
  } else {

	$winner_ref = $white_ref;
	$loser_ref = $black_ref;
  }

  my $winner_rank = $$winner_ref{"ladder_rank"};
  my $loser_rank = $$loser_ref{"ladder_rank"};

  # If the higher placed player wins nothing happens

  if ($winner_rank < $loser_rank) {

  }

  # If the lower placed player wins he moves above the higher placed player
  # Higher placed player -> Lower placed player+1 drop 1 rank

  else {

	# Write out new ranks in players.txt.tmp

	open(PLAYERS_OUT, ">go_data/players.txt.tmp") or die "can't open players.txt.tmp $!";

	my $new_rank;

	foreach my $player (values %players) {

	  if ($player == $winner_ref) {
		$new_rank = $loser_rank;
		print PLAYERS_OUT $$player{"player_name"} . "\t" . $$player{"email"} . "\t" . $$player{"strength"} . "\t" . $new_rank . "\n";
	  } elsif ($$player{"ladder_rank"} < $winner_rank && $$player{"ladder_rank"} >= $loser_rank) {
		$new_rank = $$player{"ladder_rank"} + 1;
		print PLAYERS_OUT $$player{"player_name"} . "\t" . $$player{"email"} . "\t" . $$player{"strength"} . "\t" . $new_rank . "\n";
	  } else {
		print PLAYERS_OUT $$player{"player_name"} . "\t" . $$player{"email"} . "\t" . $$player{"strength"} . "\t" . $$player{"ladder_rank"} . "\n";
	  }
	}
	  
	close(PLAYERS_OUT);
	  
	  
	# Copy new players file over old one
	  
	rename("go_data/players.txt.tmp", "go_data/players.txt") or die "can't overwrite $!";
	  
  }

  # Print HTML
	
  start_html('CUGOS Ladder Enter Game Result Succeeded!');

  print $cgi->h1("CUGOS Ladder Enter Game Result Succeeded!");
	
  print $cgi->p("Game result has been successfully added to the record");

  # No change to rankings
  if ($winner_rank < $loser_rank) {
	print $cgi->p("No change to ladder rankings!");
  } else {
	my $promoted_str = "Promoted: " . $$winner_ref{"player_name"} . " to ladder rank: ";
	my $demoted_str = "Demoted: " . $$loser_ref{"player_name"} . " to ladder rank: ";

	print $cgi->p($promoted_str, $loser_rank);
	print $cgi->p($demoted_str, $loser_rank + 1);
  }

	  
  return_main_menu_html();
	  
  print $cgi->end_html();

  close (GAMES);
}

sub enter_game_fail($) {

  my ($error_message) = @_;

  # Print failure HTML

  start_html('CUGOS Ladder Enter Game Result Failed!');
  print $cgi->h1("CUGOS Ladder Enter Game Result Failed!");
	
  print $cgi->p($error_message);
  print $cgi->p("Please re-enter data and try again.");

  print $cgi->startform("POST", "$cgi_filename");
  print $cgi->p($cgi->submit("enter_game_result_form", "Enter Game Result"));
  print $cgi->endform();

  return_main_menu_html();

  print $cgi->end_html();
}

# # lookup_player()

# # Display player details and games

sub lookup_player {

  my $player_name = $cgi->param("Player Name");

  my %players = get_players_struct();
  my @games = get_games_struct();

  # Get reference

  my $player_ref = $players{$player_name};

  # Find all games he has been in

  my @player_games;

  foreach my $game (@games) {

	if ($$game{"white"} eq $player_name || $$game{"black"} eq $player_name) {
	  push(@player_games, $game);
	}
  }

  # Print HTML

  start_html("CUGOS Ladder Player Lookup: $player_name");
  print $cgi->h1("CUGOS Ladder Player Lookup: $player_name");
	
  # player details

  my @rows = $cgi->th({-class=>'lookup_player'}, ["Player Name", "Email", "Strength", "Ladder position"]);
  push(@rows, $cgi->td({-class=>'lookup_player'}, [$$player_ref{"player_name"}, $$player_ref{"email"}, $$player_ref{"strength"}, $$player_ref{"ladder_rank"}]));

  print $cgi->table({-class=>'lookup_player'}, "", "", $cgi->Tr({-class=>'lookup_player'}, \@rows));

  print '<div class="lookup_player" />';

  # games

  my @grows = $cgi->th({-class=>'lookup_player'}, ["Date", "Black", "", "White"]);

  foreach my $game (@player_games) {

	my $result_string;

	if ($$game{"result"}) {
	  $result_string = " beat ";
	} else {
	  $result_string = " lost to ";
	}

    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($$game{"time"});
	my $time_str = "$mday" . "/" . ($mon + 1) . "/" . ($year + 1900);

	push(@grows, $cgi->td({-class=>'lookup_player'}, [$time_str, $$game{"white"}, $result_string, $$game{"black"}]));
  }

  print $cgi->table({-class=>'lookup_player'}, "", $cgi->caption({-class=>'lookup_player'}, ($player_name ."'s game record")), $cgi->Tr({-class=>'lookup_player'}, \@grows));

  return_main_menu_html();

  print $cgi->end_html();
}

sub lookup_player_form {
	
  my %players = get_players_struct();

  # build players names array

  my @player_names;

  foreach my $player_ref (values %players) {
		
	push(@player_names, $$player_ref{"player_name"});
  }

  @player_names = sort @player_names;

  # Draw lookup player form

  start_html('CUGOS Ladder Lookup Player');
  print $cgi->h1("CUGOS Ladder Lookup Player");
  print $cgi->p("Fill in the form below to get the dirt on...");
	
  print $cgi->startform("POST", "$cgi_filename");

  print $cgi->p($cgi->popup_menu(-name=>'Player Name', -values=>\@player_names));
  print $cgi->p($cgi->submit("lookup_player", "Lookup Player Data"));

  print $cgi->endform();

  return_main_menu_html();
	
  print $cgi->end_html();
}

# # modify_player()

# # Modify player details

sub modify_player {

  my $player_name = $cgi->param("Player Name");

  my %players = get_players_struct();
  my @games = get_games_struct();

  # Get reference

  my $player_ref = $players{$player_name};

  # Print HTML

  start_html("CUGOS Ladder Modify Player: $$player_ref{'player_name'}");
  print $cgi->h1("CUGOS Ladder Modify Player: $$player_ref{'player_name'}");
	
  # player details

  print $cgi->startform("POST", "$cgi_filename");

  print $cgi->p("Modify " . $player_name . "'s details...");
  
  # changing name has been disabled
#  print $cgi->p("Player name: ",
#				$cgi->textfield('Player name', $$player_ref{"player_name"}, 20, 30));

  print $cgi->p("Email (optional): ", 
				$cgi->textfield('Player email', $$player_ref{"email"}, 20, 30));

  my @go_ranks = get_go_ranks();

  print $cgi->p("Player strength: ",
				$cgi->popup_menu(-name=>'Player strength', -values=>\@go_ranks, -default=>$$player_ref{"strength"}));

  # pass original name too!
  print $cgi->hidden(-name=>'Original name',
                        -default=>$$player_ref{'player_name'});

  print $cgi->p($cgi->submit("modify_player_disc", "Modify Player"));
  
  print $cgi->endform();

  return_main_menu_html();

  print $cgi->end_html();
}

sub modify_player_disc {

  # Modify player entry on disc

  open(PLAYERS_OUT, ">go_data/players.txt.tmp") or die "can't open players.txt.tmp $!";

  my %players = get_players_struct();
  my $player_ref = $players{$cgi->param("Original name")};

  foreach my $player (values %players) {
	
	if($player == $player_ref) {
#	  print PLAYERS_OUT $cgi->param("Player name") . "\t" . $cgi->param("Player email"). "\t" . $cgi->param("Player strength") . "\t" . $$player{"ladder_rank"} . "\n";
	  print PLAYERS_OUT $$player{"player_name"} . "\t" . $cgi->param("Player email"). "\t" . $cgi->param("Player strength") . "\t" . $$player{"ladder_rank"} . "\n";
	}
	else {
	  print PLAYERS_OUT $$player{"player_name"} . "\t" . $$player{"email"} . "\t". "\t" . $$player{"strength"} . "\t" . $$player{"ladder_rank"} . "\n";
	}
  }

  close(PLAYERS_OUT);

  # Copy new players file over old one
  
  rename("go_data/players.txt.tmp", "go_data/players.txt") or die "can't overwrite $!";

  # Print HTML
	
  start_html('CUGOS Ladder Modify Player Succeeded!');

  print $cgi->h1("CUGOS Ladder Enter Game Result Succeeded!");
	
  print $cgi->p("Player record for " . $cgi->param('Original name') . " has been updated.");

  return_main_menu_html();
	
  print $cgi->end_html();

}

sub modify_player_form {
	
  my %players = get_players_struct();

  # build players names array

  my @player_names;

  foreach my $player_ref (values %players) {
		
	push(@player_names, $$player_ref{"player_name"});
  }

  @player_names = sort @player_names;

  # Draw lookup player form

  start_html('CUGOS Ladder Modify Player');
  print $cgi->h1("CUGOS Ladder Modify Player");
  print $cgi->p("Please select a player to modify...");
	
  print $cgi->startform("POST", "$cgi_filename");

  print $cgi->p($cgi->popup_menu(-name=>'Player Name', -values=>\@player_names));
  print $cgi->p($cgi->submit("modify_player", "Modify Player Data"));

  print $cgi->endform();

  return_main_menu_html();
	
  print $cgi->end_html();
}

# check_game_form()

# Calculate handicap for a game

sub check_game_form {
	
  my %players = get_players_struct();

  # build players names array

  my @player_names;

  foreach my $player_ref (values %players) {
		
	push(@player_names, $$player_ref{"player_name"});
  }

  @player_names = sort @player_names;

  # Draw check game form

  start_html('CUGOS Ladder Check Game');
  print $cgi->h1("CUGOS Ladder Check Game");
  print $cgi->p("Fill in the form below to calculate the handicap for a forthcoming game...");
	
  print $cgi->startform("POST", "$cgi_filename");

  print "<p>";
		
  print $cgi->popup_menu(-name=>'Player1',
						 -values=>\@player_names);
	
  print " verses: ";
  print $cgi->popup_menu(-name=>'Player2',
						 -values=>\@player_names);

  print "</p>";

  print $cgi->p($cgi->submit("check_game", "Check Game"));

  print $cgi->endform();

  return_main_menu_html();
	
  print $cgi->end_html();
}

sub check_game {

  my %players = get_players_struct();

  my $player1_name = $cgi->param("Player1");
  my $player2_name = $cgi->param("Player2");

  my $player1_rank = $players{$player1_name}{"strength"};
  my $player2_rank = $players{$player2_name}{"strength"};

  my $handicap = translate_rank($player2_rank) - translate_rank($player1_rank);
  my $komi;
  my $reverse_komi = 0;

  my $strong_player = $player2_name;
  my $weak_player = $player1_name;

  if ($handicap > 0) {
	$strong_player = $player1_name;
	$weak_player = $player2_name;
  }
	
  # abs and round

  $handicap = abs($handicap);

  if ($handicap > 9) {
	$komi = ($handicap - 9) * 10;
	$reverse_komi = 1;
	$handicap = 9;
  } elsif ($handicap == 0) {
	$komi = 6.5;
  }

  # Print success HTML

  start_html('CUGOS Ladder Check Game Handicap Result');
  print $cgi->h1("CUGOS Ladder Check Game Handicap Result");
	
  if ($reverse_komi) {
	print $cgi->p("$strong_player will take White and give $weak_player $handicap stones and $komi reverse komi. Good luck.");
  } elsif ($handicap) {
	print $cgi->p("$strong_player will take White and give $weak_player $handicap stones.");
  } else {
	print $cgi->p("$strong_player will play $weak_player in an even game. Komi is $komi.");
  }

  return_main_menu_html();

  print $cgi->end_html();
}



# help()

# Display help

sub help {
	
  # Print HTML

  start_html('CUGOS Ladder Help');
  print $cgi->h1("CUGOS Ladder Help");
	
  # player list

  print $cgi->startform("POST", "$cgi_filename");

  print $cgi->p("Welcome to the CUGOS Ladder system.");
  print $cgi->p("Use ",  $cgi->submit(-name=>"add_player_form", -value=>"Add Player"), " to enter your details and join the ladder at the bottom. Entering your email is optional and you may wish to obfuscate it (ie. leave off the cam.ac.uk)");
  print $cgi->p("Challenge someone up to 3 places higher than you to a game. This might take place at a regular meeting, at CB1 or on KGS.");
  print $cgi->p("Games are played on full club handicap.");
  print $cgi->p("If you win you will be promoted over your opponent. If you lose rankings are unchanged.");
  print $cgi->p("Send any questions to: <b>tom.ford (at) compsoc.net</b>. Or see me at a meeting.");

  print $cgi->endform();

  return_main_menu_html();

  print $cgi->end_html();
}



sub rank_format_data($) {

  my ($rank) = @_;

  my $printable = sprintf("%.1f", $rank);

  return $printable;
}

sub ladder_rank_format_human($) {

  my ($rank) = @_;

  my $points = 0;

  foreach my $ladder_pos (@ladder_ranks) {

	$points++;

	if (abs($ladder_pos - $rank) < $epsilon) {
	  last;
	}
  }

  my $printable = $points;
	
  return $printable;
}

sub strength_rank_format_human($) {

  my ($rank) = @_;

  my $printable = $rank;

  if ($printable > 0) {
	$printable = $printable . " kyu";
  } else {
	$printable = (-1 * $printable) + 1;
	$printable = $printable . " dan";
  }
	
  return $printable;
}

sub start_html($) {
  
  my ($title) = @_;
  
  my %style_hash = ( 'src' => '../gostyle.css' );

  print $cgi->header();
  print $cgi->start_html(-style=>\%style_hash, -title=>$title);
}

sub return_main_menu_html() {

  # Return to main menu footer

  print '<div class="main_menu_footer">';

  print $cgi->startform("POST", "$cgi_filename");
  print $cgi->submit(-class=>'main_menu_footer', -name=>"main_menu", -value=>"Main Menu");
  print $cgi->endform();

  print "</div>"
}

sub get_go_ranks() {

  my @go_ranks;

  for (my $i = 30; $i >= 1; $i--) {
	push(@go_ranks, "$i kyu");
  }

  for (my $i = 1; $i <= 7; $i++) {
	push(@go_ranks, "$i dan");
  }
  
  return @go_ranks;
}

sub translate_rank($) {
  
  my ($input_rank) = @_;

  $input_rank =~ /^(\d+)\ (\w+)/;

  if ($2 eq "dan") {
	return (-1 * $1) + 1;
  }
  return $1;
}
